/**
 * @author Nicolas Limare <nicolas.limare@cmla.ens-cachan.fr>
 */

/**
 * less interesting tanh() approximations
 *
 * tanh_rational: fastest, but only "looks like" tanh(), can be
 * useful to compress a signal, for example, but not when we want
 * something close to the mathematical definition of tanh()
 *
 * tanh_pade: smooth polynomial, close to tanh() but slow
 *
 * tanh_composite: high precision but slow, probably close to the
 * glibc tanh()
 */

#ifndef _TANH_EXTRA_HPP_
#define  _TANH_EXTRA_HPP_

/* selectively allow unused functions and variables */

#ifndef _UNUSED
#if defined(__GNUC__)
/* from http://sourceforge.net/p/predef/wiki/Compilers/ */
#define _UNUSED __attribute__((__unused__))
#else
#define _UNUSED
#endif
#endif

/**
 * tanh() approximation by a clipped rational
 *
 * return -1 if x < -3,
 *         1 if x >  3,
 *          otherwise x * (27 x^2) / (27 + 9 x^2)
 *
 * Takes 4 products, 2 sums, 1 division and some test branching
 * The two first derivatives vanish at -3 and 3, this approximation
 * C2-continuous.
 *
 * from http://www.musicdsp.org/showone.php?id=238
 */
_UNUSED
static float tanh_rational(float x)
{
    float sign = (float) ((x > 0) - (x < 0));
    return (fabsf(x) < 3.
            ? x * (27. + x * x) / (27. + 9. * x * x)
            : sign);
}

/**
 * tanh_rational if x is known to always be between -3 and 3
 *
 * Without the test, it's slightly faster
 */
_UNUSED
static float tanh_rational_notest(float x) {
    return (x * (27. + x * x) / (27. + 9. * x * x));
}

/**
 * tanh() pade approximation
 *
 * return x (600 + 70 x^2 + x^4)
 *          / (600 + 270 x^2 + 11 x^4 + x^6 / 24)
 */
_UNUSED
static float tanh_pade(float x)
{
    return (x * (600. + x*x * (70. + x*x))
            / (600. + x*x * (270. + x*x * (11. + 1./24.))));
}

/**
 * tanh() composite approximation
 *
 * return x                      if x < xsmall
 *        a rational poly        if x < xlarge
 *        1 - 2 / ( 1 + exp(2x)) if x < xlarge
 *        1                      else
 *
 * From Nelson H. F. Beebe's work
 *   http://www.math.utah.edu/~beebe/software/ieee/tanh.pdf
 *   http://www.math.utah.edu/~beebe/software/ieee/dtanh.f
 *
 */

_UNUSED
static float tanh_composite(float x)
{
    const float xlarge  = 19.0615474653E+00;
    const float xmedium =  0.5493061443E+00;
    const float xsmall  =  1.2904784139E-08;

    const double p0 = -0.8237728127E+00;
    const double p1 = -0.3831010665E-02;
    const double q0 =  0.2471319654E+01;
    const double q1 =  1.00000000000E+00;

    float _x = fabs(x);
    float retval;

#pragma GCC diagnostic ignored "-Wfloat-equal"
    if (_x != _x)
        // x is NaN, generate a run-time trap
        retval = _x + _x;
    else if (_x >= xlarge)
        retval = 1.;
    else if (_x >= xmedium)
        retval = 1 - 2 / (exp(_x + _x) + 1);
    else if (_x >= xsmall) {
        // FIXME: there is a numerical eror here, inconsistent results
        double xx = _x * _x;
        double p = p1 * xx + p0;
        double q = q1 * xx + q0;
        retval = xx * p / q;
    }
    else
        retval = _x;
    if (x < 0)
        retval = -retval;
#pragma GCC diagnostic pop

    return retval;
}

#endif  /* !_TANH_EXTRA_HPP_ */
